#!/usr/bin/env python3
##import dependency libraries
import sys,time,glob,os,pickle,fileinput,argparse
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import seaborn as sns
import pandas as pd
from scipy.stats import rankdata
from matplotlib.colors import LinearSegmentedColormap
try:
    from pyminer.common_functions import *
except:
    from common_functions import *
# ##############################################################
# ## basic function library
# def read_file(tempFile,linesOraw='lines',quiet=False):
#     if not quiet:
#         print('reading',tempFile)
#     f=open(tempFile,'r')
#     if linesOraw=='lines':
#         lines=f.readlines()
#         for i in range(0,len(lines)):
#             lines[i]=lines[i].strip('\n')
#     elif linesOraw=='raw':
#         lines=f.read()
#     f.close()
#     return(lines)

# def make_file(contents,path):
#     f=open(path,'w')
#     if isinstance(contents,list):
#         f.writelines(contents)
#     elif isinstance(contents,str):
#         f.write(contents)
#     f.close()

    
# def flatten_2D_table(table,delim):
#     #print(type(table))
#     if str(type(table))=="<class 'numpy.ndarray'>":
#         out=[]
#         for i in range(0,len(table)):
#             out.append([])
#             for j in range(0,len(table[i])):
#                 try:
#                     str(table[i][j])
#                 except:
#                     print(table[i][j])
#                 else:
#                     out[i].append(str(table[i][j]))
#             out[i]=delim.join(out[i])+'\n'
#         return(out)
#     else:
#         for i in range(0,len(table)):
#             for j in range(0,len(table[i])):
#                 try:
#                     str(table[i][j])
#                 except:
#                     print(table[i][j])
#                 else:
#                     table[i][j]=str(table[i][j])
#             table[i]=delim.join(table[i])+'\n'
#     #print(table[0])
#         return(table)

# def strip_split(line, delim = '\t'):
#     return(line.strip('\n').split(delim))

# def make_table(lines,delim):
#     for i in range(0,len(lines)):
#         lines[i]=lines[i].strip()
#         lines[i]=lines[i].split(delim)
#         for j in range(0,len(lines[i])):
#             try:
#                 float(lines[i][j])
#             except:
#                 lines[i][j]=lines[i][j].replace('"','')
#             else:
#                 lines[i][j]=float(lines[i][j])
#     return(lines)


# def get_file_path(in_path):
#     in_path = in_path.split('/')
#     in_path = in_path[:-1]
#     in_path = '/'.join(in_path)
#     return(in_path+'/')


# def read_table(file, sep='\t'):
#     return(make_table(read_file(file,'lines'),sep))
    
# def write_table(table, out_file, sep = '\t'):
#     make_file(flatten_2D_table(table,sep), out_file)
    

# def import_dict(f):
#     f=open(f,'rb')
#     d=pickle.load(f)
#     f.close()
#     return(d)

# def save_dict(d,path):
#     f=open(path,'wb')
#     pickle.dump(d,f)
#     f.close()

# def cmd(in_message, com=True):
#     print(in_message)
#     time.sleep(.25)
#     if com:
#         Popen(in_message,shell=True).communicate()
#     else:
#         Popen(in_message,shell=True)



# ##############################################################












import argparse
parser = argparse.ArgumentParser()

parser.add_argument(
	'-infile','-in','-i','-input',
	dest='infile',
	type=str)

parser.add_argument("-nodes_of_interest", "-noi","-nodes",
    help="the nodes that you want to subset and get their degrees of separation from")

parser.add_argument("-out_dir", "-o", "-out",
    help="the directory to place the output file(s)")

parser.add_argument("-cluster_pkl", 
    help="the file created by clustering.py that stores the plots")

parser.add_argument(
    '-no_cluster','-no_clust',
    help = "if you don't want the heatmap to cluster the genes together, but rather, just want it to be in the order that you input them.",
    dest = "cluster",
    default = True,
    action = 'store_false')
parser.add_argument(
    '-only_heatmap','-heatmap',
    help = "if you only want the heatmap",
    dest = "only_heatmap",
    default = False,
    action = 'store_true')
parser.add_argument(
    '-no_rank',
    help = "if you don't want the heatmap rank transformed",
    default = False,
    action = 'store_true')

## hdf5 options
parser.add_argument(
    '-hdf5',
    help = 'The input file is an HDF5 file',
    default = False,
    action = 'store_true')

parser.add_argument(
    "-ID_list","-ids","-id_list",
    help = 'If we are using an hdf5 file, give the row-wise IDs in this new line delimeted file',
    type = str)

parser.add_argument(
    "-columns","-cols",
    help = 'If we are using an hdf5 file, give the column-wise IDs in this new line delimeted file',
    type = str)




args = parser.parse_args()
##############################################################

def ids_to_idxs(in_ids):
    global ID_hash
    temp_index_list = [ID_hash[i] for i in in_ids]
    return(temp_index_list)


def process_noi(nodes_of_interest):
    #check if there are aliases
    noi_table=[]
    for i in range(0,len(nodes_of_interest)):
        if '\t' in nodes_of_interest[i]:
            noi_table.append(nodes_of_interest[i].split('\t'))
        elif '::' in nodes_of_interest[i]:
            noi_table.append(nodes_of_interest[i].split('::'))
        else:
            noi_table.append([nodes_of_interest[i],nodes_of_interest[i]])

    # print(noi_table)
    ## also make the ailias look-up dictionary
    alias_dict={}
    ## make the string that would be passed through to command line if needed
    pass_through_arg =""
    ## and the final list of nodes in the dataset
    final_nodes = []
    for i in range(0,len(noi_table)):
        # print(i)
        # print(noi_table[i])
        alias_dict[noi_table[i][0]]=noi_table[i][1]
        pass_through_arg+=","+noi_table[i][0]+'::'+noi_table[i][1]
        final_nodes.append(noi_table[i][0])
    ## remove the leading comma
    pass_through_arg=pass_through_arg[1:]

    return(alias_dict, pass_through_arg, final_nodes)
##########################
######################################


def linear_normalize(vect):
    vect = vect - min(vect)
    vect = vect/max(vect)
    return(vect)


def vect_to_color(vect,style='hot'):
    vect = linear_normalize(vect)
    cmap = cm.get_cmap(name=style)
    cmap = LinearSegmentedColormap.from_list("WhRd", ((1,1,1),(1,0,0)), N=256)
    return(cmap(vect))

def draw_plot(plot,exemplars,color_vect, colorize_by_gene = None):
    global all_group_ids, plt, full_expression, ID_list, ID_hash
    plt.clf()

    ## get the colorized vector if need be
    if colorize_by_gene!=None:
        ## fix for entrez..
        try:
            float(colorize_by_gene)
        except:
            colorize_by_gene = colorize_by_gene
        else:
            colorize_by_gene = str(float(colorize_by_gene))
        expression_vect = full_expression[ID_hash[colorize_by_gene],:]
        expression_vect = rankdata(expression_vect,method='min')
        colorized_vect = vect_to_color(expression_vect,style='Reds')

    if colorize_by_gene!=None:
        plt.scatter(plot["x"],plot["y"],
                edgecolor='k',#plot["c"],
                facecolor=colorized_vect,
                s=10,
                linewidths=0.075)
    else:
        plt.scatter(plot["x"],plot["y"],
                color=plot["c"],
                s=10,
                linewidths=0.075)
    plt.xlabel(plot['xlab'])
    plt.ylabel(plot['ylab'])




def lin_norm_rows(in_mat,min_range=0,max_range=1):
    in_mat = np.transpose(np.array(in_mat))
    in_mat = in_mat - np.min(in_mat, axis = 0)
    in_mat = in_mat / np.max(in_mat, axis = 0)
    in_mat[np.isnan(in_mat)]=0
    return(np.transpose(in_mat))


def names_to_alias(names):
    global alias_dict
    aliases=[]
    for i in range(0,len(names)):
        ## fix for entrez..
        try:
            float(names[i])
        except:
            temp_id = names[i]
        else:
            temp_id = str(int(float(names[i])))
        aliases.append(alias_dict[temp_id])
    return(aliases)



#######################################

if args.nodes_of_interest != None:
	## check if it's a file
	if os.path.isfile(args.nodes_of_interest):
		nodes_of_interest = read_file(args.nodes_of_interest,'lines')
	else:
		## check if it's a comma separated list of ids
		nodes_of_interest = args.nodes_of_interest.split(',')
		# nodes_of_interest = args.nodes_of_interest
		# nodes_of_interest = nodes_of_interest.split(',')
else:
    nodes_of_interest = None



alias_dict, pass_through_arg, temp_nodes_of_interest = process_noi(nodes_of_interest)
print(pass_through_arg)


if args.ID_list!=None:
	ID_list = read_file(args.ID_list, "lines")
	ID_hash = {}
	for i in range(0,len(ID_list)):
		ID_hash[ID_list[i]] = i
else:
	sys.exit('-ID_list </path/to/file> is a required argument')

if args.out_dir != None:
	out_dir = args.out_dir
else:
	sys.exit('-out_dir is a required argument')

print(ID_list[:5])
## double check that the ids are found
nodes_of_interest=[]
for i in range(0,len(temp_nodes_of_interest)):
    ## fix for entrez..
    try:
        float(temp_nodes_of_interest[i])
    except:
        temp_id = temp_nodes_of_interest[i]
    else:
        temp_id = str(float(temp_nodes_of_interest[i]))
    if temp_id in ID_hash:
        nodes_of_interest.append(temp_id)
    else:
        print("couldn't find",temp_id)

if len(nodes_of_interest)==0:
    sys.exit("couldn't find any of the desired genes")

## import expression data
if not args.hdf5:
    full_expression_str = read_table(args.infile)
    title = full_expression_str[0]
    full_expression_np = np.array(full_expression_str)
    row_names = full_expression_np[1:,0]
    full_expression = np.array(full_expression_np[1:,1:],dtype = float)
else:
    row_names = read_file(args.ID_list,'lines')
    title = read_file(args.columns,'lines')
    print('making a maliable hdf5 file to preserve the original data')
    cmd('cp '+args.infile+' '+args.infile+'_copy')
    import h5py
    print('reading in hdf5 file')
    infile_path = args.infile+'_copy'
    h5f = h5py.File(infile_path, 'r+')
    full_expression=h5f["infile"]

ID_list = row_names[:]
ID_hash = {name:idx for idx,name  in enumerate(ID_list)}


############################################




# single_cell_type_table = read_table(args.single_groups,num_type=int)
# single_group_lists = reorganize_single_type_table(single_cell_type_table)

#############################################



## read in the plots
cluster_result_dict = import_dict(args.cluster_pkl)
print(list(cluster_result_dict.keys()))

linear_groups = cluster_result_dict['linear_groups']
sample_k_lists = cluster_result_dict["sample_k_lists"]


##################################
args.out_dir = process_dir(args.out_dir)

# if args.out_dir[:-1]!= '/':
# 	args.out_dir+='/'
##################################


## plot the heatmap
group_reordering_vector=np.array(cluster_result_dict['group_reordering_vector'])
reordered_colors=np.array(cluster_result_dict['reordered_colors'])


#######################################################
## get the column labels

reordered_array = list(np.array(linear_groups)[np.array(group_reordering_vector)])
#print(reordered_array)
reordered_groups = [reordered_array[0]]
#print(reordered_groups)
for i in range(1,len(reordered_array)):
	#print(int(reordered_array[i]))
	#print(reordered_groups[-1])
	if int(reordered_array[i])!=int(reordered_groups[-1]):
		reordered_groups.append(reordered_array[i])

#print(reordered_groups)

cluster_lists = sample_k_lists[:]
temp_label_lists = []
for i in range(0,len(cluster_lists)):
	temp_list = ['']*len(cluster_lists[i])
	temp_list[int(len(temp_list)/2)] = i
	temp_label_lists.append(temp_list)

heatmap_x_labels = []
for i in reordered_groups:
	heatmap_x_labels+=temp_label_lists[i]
#print(heatmap_x_labels)

#######################################################

# print(group_reordering_vector)
# print(reordered_colors)
# sys.exit()
original_order = ids_to_idxs(nodes_of_interest)
# from scipy import stats
# original_order_rank = np.array(stats.rankdata(original_order)-1,dtype=int)
# print(original_order_rank)
original_order_dict = {j:i for i,j in enumerate(original_order)}


subset_indices = sorted(original_order)
original_order_vector = []
for i in subset_indices:
    ## fix for entrez..
    try:
        float(ID_list[i])
    except:
        temp_id = ID_list[i]
    else:
        temp_id = str(int(float(ID_list[i])))
    print(i,original_order_dict[i], temp_id, alias_dict[temp_id])
    original_order_vector.append(original_order_dict[i])

new_order_dict = {i:j for i,j in enumerate(original_order_vector)}
original_order_rank=np.zeros((len(original_order_vector)),dtype=int)
for i in range(0,len(original_order_vector)):
    index = original_order_vector[i]
    original_order_rank[index]=i
original_order_rank=original_order_rank.tolist()


original_order_names = np.array(ID_list)[original_order]
subset_sorted_names = np.array(ID_list)[subset_indices]
subset_indices = np.array(subset_indices)
print(subset_indices)
if np.shape(subset_indices)[0]==0:
    sys.exit('no subset found...')
subset_array = np.zeros((subset_indices.shape[0],full_expression.shape[1]))
for i in range(subset_indices.shape[0]):
    temp_idx = subset_indices[i]
    subset_array[i,:] = np.array(full_expression[temp_idx,:])


# if not args.cluster:
#     ## put them back in the input order
#     subset_array=subset_array[original_order_rank,:]
#     subset_sorted_names = subset_sorted_names[original_order_rank]

## do the linear normalization
subset_array = lin_norm_rows(subset_array[:,group_reordering_vector])

## Z-score
# import scipy
# subset_array = scipy.stats.zscore(subset_array,axis=1)
# subset_array = lin_norm_rows(subset_array)
# ## mean center
#subset_array = subset_array - np.transpose(np.array([np.mean(subset_array,axis=1)]))

if not args.no_rank:
    from scipy.stats import rankdata
    for i in range(0,np.shape(subset_array)[0]):
        subset_array[i] = rankdata(subset_array[i,:],method='min')
else:
    ## if we're not doing rank, do Z-score
    from scipy.stats import zscore
    subset_array = zscore(subset_array, axis = 1)
    subset_array[subset_array>3] = 3
    subset_array[subset_array<-3] = -3
#subset_array = np.log2(subset_array)


name_col = np.array(names_to_alias(subset_sorted_names))
print(name_col)
name_col_df = pd.DataFrame({"name_col":name_col},index=list(range(len(subset_sorted_names))))
#subset_df = np.hstack((name_col, subset_array))
subset_df = pd.DataFrame(data=subset_array)
subset_df = pd.concat([name_col_df,subset_df],axis=1)
print(subset_df)
#subset_df = subset_df.set_index([list(range(len(subset_sorted_names))),"name_col"])
subset_df = subset_df.set_index([list(original_order_rank),"name_col"])
subset_df.columns = heatmap_x_labels
print(subset_df)

## set some stuff around if we are or are not clustering the rows
if not args.cluster:
    print(original_order_rank)
    orig_subset_df = subset_df
    #subset_df=subset_df.reindex(original_order_rank)
    subset_df = subset_df.iloc[original_order_rank,:]
    print(subset_df)
    row_cluster = False
else:
    row_cluster = True

print(subset_df)

#print(subset_df[:5,:5])

subset_sorted_aliases = names_to_alias(subset_sorted_names)
if not args.cluster:
    original_order_aliases = names_to_alias(original_order_names)
    for i in (zip(original_order_names,original_order_aliases)):
        print(i)

try_reorder=False

if args.no_rank:
    global_cmap = sns.color_palette("coolwarm", 256)
else:
    global_cmap = cm.get_cmap(name="hot")

## actually plot the heatmap finally...
## get the non-empty colnames
indices_to_plot_for_names = np.where(np.array(heatmap_x_labels)!='')[0]
#indices_to_plot_for_names = np.where(np.array(heatmap_x_labels)=='')[0]
none_indices = np.where(np.array(heatmap_x_labels)=='')[0].tolist()
#print(none_indices[:5])
plot_names = np.array(heatmap_x_labels)
plot_names[none_indices]=None
print(indices_to_plot_for_names)
xticklabels = False#[None]*len(heatmap_x_labels)
try:
    sub = sns.clustermap(subset_df,
        col_cluster = False,
        row_cluster = row_cluster,
        col_colors = reordered_colors,
        yticklabels = True,
        #xticklabels=xticklabels,
        cmap=global_cmap)
except:
    ## if it were too big to plot successfully, just do a subset
    sample = np.arange((np.shape(subset_df)[0]))
    subset_length = min([np.shape(subset_df)[0],400])
    np.random.shuffle(sample)
    sub_sample = np.sort(sample[:subset_length])
    sub = sns.clustermap(np.array(subset_df)[sub_sample,:],
        col_cluster = False,
        row_cluster = row_cluster,
        col_colors = reordered_colors,
        yticklabels = True,
        #xticklabels=xticklabels,
        cmap=global_cmap)

## this sets the heatmap labels on the x axis
sub.ax_heatmap.xaxis.set_ticks(indices_to_plot_for_names)
sub.ax_heatmap.xaxis.set_ticklabels(plot_names[indices_to_plot_for_names])

plt.setp(sub.ax_heatmap.yaxis.get_majorticklabels(), rotation=0)
plt.setp(sub.ax_heatmap.xaxis.get_majorticklabels(), rotation=90)

# plt.tight_layout()
# plt.show()
plt.savefig(args.out_dir+'genes_of_interest_subset_heatmap.png',
    dpi=600,
    bbox_inches='tight')

if args.only_heatmap:
    sys.exit()

#sys.exit()
# ## plot the nodes of interest on the tSNE and PCAs
for noi in nodes_of_interest:
    ## fix for entrez..
    try:
        float(noi)
    except:
        noi = noi
    else:
        noi = str(int(float(noi)))
    for plot in cluster_result_dict["plots"]:
        # for k in [plot].keys():
        #     print(k)
        #     print(plot[k])
        draw_plot(cluster_result_dict["plots"][plot],cluster_result_dict["exemplar_indices"],cluster_result_dict["color_vect"],colorize_by_gene = noi)
        temp_plot_out = plot.split('/')
        temp_plot_out = temp_plot_out[-1]
        plt.savefig(args.out_dir+temp_plot_out[:-4]+'_'+alias_dict[noi]+'.png',
            dpi=600,
            bbox_inches='tight')





#!/usr/bin/env python3
##import dependency libraries
import sys,time,glob,os,pickle,fileinput, random, h5py
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
import numpy as np
from statsmodels.nonparametric.smoothers_lowess import lowess
from sklearn.neighbors import RadiusNeighborsRegressor as neighbors

import matplotlib.pyplot as plt
##############################################################
## basic function library
def read_file(tempFile,linesOraw='lines',quiet=False):
    if not quiet:
        print('reading',tempFile)
    f=open(tempFile,'r')
    if linesOraw=='lines':
        lines=f.readlines()
        for i in range(0,len(lines)):
            lines[i]=lines[i].strip('\n')
    elif linesOraw=='raw':
        lines=f.read()
    f.close()
    return(lines)

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

    
def flatten_2D_table(table,delim):
    #print(type(table))
    if str(type(table))=="<class 'numpy.ndarray'>":
        out=[]
        for i in range(0,len(table)):
            out.append([])
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i]=delim.join(out[i])+'\n'
        return(out)
    else:
        for i in range(0,len(table)):
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j]=str(table[i][j])
            table[i]=delim.join(str(table[i]))+'\n'
    #print(table[0])
        return(table)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))

def make_table(lines,delim):
    for i in range(0,len(lines)):
        lines[i]=lines[i].strip()
        lines[i]=lines[i].split(delim)
        for j in range(0,len(lines[i])):
            try:
                float(lines[i][j])
            except:
                lines[i][j]=lines[i][j].replace('"','')
            else:
                lines[i][j]=float(lines[i][j])
    return(lines)


def read_table(file, sep='\t'):
    return(make_table(read_file(file,'lines'),sep))
    
def write_table(table, out_file, sep = '\t'):
    make_file(flatten_2D_table(table,sep), out_file)
    

def import_dict(f):
    f=open(f,'rb')
    d=pickle.load(f)
    f.close()
    return(d)

def save_dict(d,path):
    f=open(path,'wb')
    pickle.dump(d,f)
    f.close()

def cmd(in_message, com=True):
    print(in_message)
    time.sleep(.25)
    if com:
        Popen(in_message,shell=True).communicate()
    else:
        Popen(in_message,shell=True)



##############################################################



##############################################################
import argparse
parser = argparse.ArgumentParser()

parser.add_argument("-infile","-i", 
    help="the input expression matrix",
    type = str)
parser.add_argument("-out_dir", '-out','-o',
    help="the directory for output files",
    type = str)
parser.add_argument("-log", 
    help="log transform the means",
    action = 'store_true',
    default = False)
parser.add_argument("-hdf5", 
    help="the input file is an hdf5 file",
    action = 'store_true',
    default = False)
parser.add_argument("-ids", '-ID_list','-id_list',
    help="if an hdf5 file is used, we'll need the list of IDs",
    type = str)
parser.add_argument("-z_cutoff", "-z",
    help="the cutoff for how many standard deviations over the mean residual should be considered overdispersed. Default = 0.5",
    type = float,
    default = .5)
args = parser.parse_args()
if args.out_dir[-1]!='/':
	args.out_dir+='/'
if not os.path.isdir(args.out_dir):
    cmd('mkdir '+args.out_dir)
##############################################################
gc.enable()
## read in the input file
if not args.hdf5:
    in_mat_str = np.array(read_table(args.infile))
    in_mat = np.array(in_mat_str[1:,1:],dtype=float)
    ID_list = in_mat_str[1:,0].tolist()
    del in_mat_str
else:
    ## read in the hdf5 file
    print('reading hdf5 file:\n\t',args.infile)
    h5f = h5py.File(args.infile, 'r')
    in_mat=h5f["infile"]
    ID_list = read_file(args.ids,'lines')




if args.log:
    if np.min(in_mat)<0.0:
        print("can't log transform with negative values... Bringing the matrix up to min = 1")
        in_mat+=min(in_mat)
    in_mat = np.log2(1 + in_mat)



## calculate the means
means = np.mean(in_mat,axis=1)
## calculate the coefficient of variance
variance = np.std(in_mat,axis=1)#/(means+1)
## square the coefficient of variance
variance*=variance

## remove the nans

print(np.sum(np.isnan(variance)))
variance[np.isnan(variance)]=0
print(np.sum(np.isnan(variance)))


linear_dim = np.shape(means)


## plot (a random sample of) the relationship between the coefficient of variance and means
max_sample_size = 50000
sample_size = min([max_sample_size,linear_dim[0]])
if sample_size < max_sample_size:
	print('sampling the distribution for',sample_size,'points')
	index_vect  = np.arange(linear_dim[0])
	np.random.shuffle(index_vect)
	sample = index_vect[:sample_size].tolist()
	## always include the min and max values for interpolation
	x_max_index = np.where(np.array(means) == np.max(means))[0][0]
	x_min_index = np.where(np.array(means) == np.min(means))[0][0]
	#print(sample)
	#print(type(sample))
	#sample = sample.tolist()
	if x_max_index not in sample:
		sample.append(x_max_index)
	if x_min_index not in sample:
		sample.append(x_min_index)
	#print(sample)
	print(type(sample))
else:
    sample = np.arange(linear_dim[0], dtype =  int).tolist()
    #print(sample)
    #sys.exit()

sample = np.sort(sample)
## remove the indices where the variance is zero
keep_sample = np.where(variance[sample]>0)[0]
sample = sample[keep_sample]

plt.scatter(means[sample], variance[sample], c = 'black', s = 0.5)#,xlab = 'mean',ylab = 'CV')
#plt.show()
#sys.exit()

## do the lowess fit. This returns the expected values of the variance
print('calculating the best fit curve')

## 

neigh = neighbors(radius = 1.0, weights = 'uniform', leaf_size = 30)
mean_max = max(means)
mean_min = min(means)
sds_min = min(variance)

## first calculate the lowess curve on the sample
lowess_estimates_sample = lowess(variance[sample]-sds_min+1,means[sample]-mean_min+1, delta = 0.01*mean_max)-1
# lowess_estimates_sample[:,0] = lowess_estimates_sample[:,0] + mean_min
# lowess_estimates_sample[:,1] = lowess_estimates_sample[:,1] + sds_min

## remove the nans 
lowess_estimates_sample[np.isnan(lowess_estimates_sample)]=0

print(lowess_estimates_sample)

## remove x ties from the lowess as this messes up the interpolation
values, indices = np.unique(lowess_estimates_sample[:,0], return_index=True)
lowess_estimates_sample = np.array(lowess_estimates_sample[indices,:])
print(lowess_estimates_sample)
print(type(lowess_estimates_sample))
#sys.exit()

## then interpolate the curve to smooth it out and provide the nearest neighbor regression with the needed coverage
from scipy.interpolate import interp1d

# ## positive control
# x = np.linspace(0, 10, num=11, endpoint=True)
# print(x)
# y = np.cos(-x**2/9.0)
# print(y)
# f2 = interp1d(x, y, kind='cubic')
# print(f2)
# xnew = np.linspace(0, 10, num=41, endpoint=True)
# print(xnew)
# print(f2(xnew))

print(np.shape(lowess_estimates_sample[:,0]))
print(np.shape(lowess_estimates_sample[:,1]))

## cross the spread
print("original_min_max:",min(lowess_estimates_sample[:,0]),max(lowess_estimates_sample[:,0]))


epsilon = (max(lowess_estimates_sample[:,0]) - min(lowess_estimates_sample[:,0]))*0.01
print('epsilon:',epsilon)

#make the new x values evenly spaced a
temp_min = min(lowess_estimates_sample[:,0])# - epsilon
temp_max = max(lowess_estimates_sample[:,0])# + epsilon
new_x = np.linspace(temp_min,temp_max, num=sample_size, endpoint = True)
#new_x = np.linspace(temp_min,temp_max, num=sample_size, endpoint = True)
#new_x[0] = temp_min
#new_x[-1] = temp_max

interpolation = interp1d(lowess_estimates_sample[:,0], lowess_estimates_sample[:,1], kind='cubic')


print("new_min_max:",min(new_x),max(new_x))
print(new_x)

## interpolate the lowess function
new_y = interpolation(new_x)
print(new_x,new_y)

## get the interplated lowess ready for nearest neighbor regression
train_x = np.array([[x] for x in lowess_estimates_sample[:,0].tolist()])
train_y = lowess_estimates_sample[:,1]

full_x = np.array([[x] for x in means.tolist()])
#train_x = full_x[sample]
#train_y = variance[sample]

print(train_x)
print(train_y)
#neigh.fit(train_x,train_y)
print(new_x)
print(new_y)
new_x = np.array([[x] for x in new_x.tolist()])

#plt.clf()
plt.scatter(new_x,new_y, c = 'red', s = 0.75)
plt.savefig(args.out_dir+'local_fit.png',dpi=600,bbox_inches='tight')
#plt.show()
#sys.exit()
neigh.fit(new_x,new_y)

## now calculate the values for everything else
bin_size = 100000
total_vars = len(means)
bins = []
cur_bin = 0
while cur_bin<total_vars:
	bins.append(min(cur_bin, total_vars))
	cur_bin+=bin_size

bins.append(total_vars)
print(bins)

lowess_estimates = np.zeros((len(means),))
# for i in range(1,len(bins)):
# 	print("working on",bins[i-1],bins[i])
# 	# if i%50000 == 0:
# 	# 	print('\t',bins[i]/len(means))
# 	lowess_estimates[bins[i-1]:bins[i]] += neigh.predict(full_x[bins[i-1]:bins[i]])

for i in range(0,total_vars):
    x_dif_vect = np.abs(full_x[i]-new_x)
    closest_idx = int(np.argmin(x_dif_vect))
    lowess_estimates[i]=new_y[closest_idx]
    



#lowess_estimates = neigh.predict(full_x)

#lowess_estimates = lowess(variance[sample]+1,means[sample]+1)-1#, return_sorted = False)#, frac = 1e-3, delta = 1e-10*max(means))
#lowess_estimates = [x[0] for x in lowess_estimates]


#plt.scatter(means[sample],lowess_estimates[sample],c='red',s=0.5)
#plt.savefig(args.out_dir+'local_fit_sample.png',dpi=600,bbox_inches='tight')
#plt.show()
#sys.exit()
## calculate the residuals
print('getting the residuals')
residuals = variance - lowess_estimates


# print(lowess_estimates)
# # sys.exit()
# print(residuals - variance)
# sys.exit()
print('sample_means',means[:10])
print('sample_estimates',lowess_estimates[:10])
print('sample_CVs',variance[:10])
print('sample_residuals',residuals[:10])


## reshape it back into the table
resid_table = residuals

print(np.sum(np.isnan(residuals)))
nan_idx = np.where(np.isnan(residuals))[0]

print('nan means',means[nan_idx])
print('nan sds',variance[nan_idx])
print('nan estimates',lowess_estimates[nan_idx])
print('nan resid',residuals[nan_idx])


#print(np.nan_std(residuals))
st_dev_resid = np.std(residuals)
print('sd residuals:',st_dev_resid)

st_dev_cutoff = st_dev_resid*args.z_cutoff
print('residual cutoff for overdispersion:',st_dev_cutoff)

## plot the residuals
plt.clf()
plt.scatter(means[sample],residuals[sample], c = 'black', s = 0.5)#,xlab = 'mean',ylab = 'CV')
plt.plot([min(means),max(means)],[0,0],c='red')
## find the significant residuals
resid_sample = residuals[sample]
mean_sample = means[sample]
sig_indices = np.where(resid_sample >= st_dev_cutoff)[0]
plt.scatter(mean_sample[sig_indices], resid_sample[sig_indices], c = 'blue', s = 2)
plt.savefig(args.out_dir+'residuals.png',dpi=600,bbox_inches='tight')
#plt.show()


## determine which genes are locally overdispersed
print('calculating local overdispersion')
local_overdispersion_bool_table = resid_table >= st_dev_cutoff

print('number of overdispersed genes:')
number_overdispersed_per_group = np.sum(local_overdispersion_bool_table, axis = 0)
print(number_overdispersed_per_group)



## get the boolean overdispersed table ready for writing to file
#print(np.shape(local_overdispersion_bool_table))
#print(local_overdispersion_bool_table)
local_overdispersion_bool_table = np.array(local_overdispersion_bool_table,dtype = str)
#print(local_overdispersion_bool_table)
local_overdispersion_bool_table=local_overdispersion_bool_table.tolist()
print(local_overdispersion_bool_table[:5],len(local_overdispersion_bool_table))
print(np.shape(means)[0])
## add the gene names
gene_names = ID_list
for i in range(0,(np.shape(means)[0])):
    #print(i)
    local_overdispersion_bool_table[i] = [gene_names[i]] + [local_overdispersion_bool_table[i]]


local_overdispersion_bool_table=np.array(local_overdispersion_bool_table, dtype = str)
print(local_overdispersion_bool_table)
print(np.shape(local_overdispersion_bool_table))

write_table(local_overdispersion_bool_table,args.out_dir+'/locally_overdispersed_boolean_table.txt')







#!/usr/local/env python3

import argparse
try:
    from pyminer.common_functions import *
    from pyminer.pyminer_process_tad_bed import *
    from pyminer.pyminer_ensembl_rest import map_variants_to_genes
    from pyminer.pyminer_gene_term_comentions import get_comention_summary
    from pyminer.pyminer_gprofiler_converter import convert_to_ensg
except:
    from common_functions import *
    from pyminer_process_tad_bed import *
    from pyminer_ensembl_rest import map_variants_to_genes
    from pyminer_gene_term_comentions import get_comention_summary
    from pyminer_gprofiler_converter import convert_to_ensg
##############################################################


##############################################################

def convert_anno_file_to_ensg_lookup(annotation_file, id_list):
    anno_table = np.array(read_table(annotation_file))
    orig_indices = list(map(int,map(float,anno_table[1:,0].tolist())))
    orig_indices = [idx-1 for idx in orig_indices]## gprofiler's indexing is from 1
    #print(orig_indices)
    ensg_ids = anno_table[1:,3].tolist()
    ensg_to_original_id = {temp_id:[] for temp_id in ensg_ids}
    for i in range(len(ensg_ids)):
        temp_ensg = ensg_ids[i]
        temp_orig_list = ensg_to_original_id[temp_ensg]
        temp_orig_list.append(id_list[orig_indices[i]])
        ensg_to_original_id[temp_ensg] = temp_orig_list
    return(ensg_to_original_id)


def map_variants_to_original_ids(var_to_ensg, ensg_to_original_id):

    return()


def get_gene_level_info(all_variants,out_dir):
    genes = []
    for var in all_variants:
        genes += var.genes
    genes = sorted(list(set(genes)))
    #annotation_out_path = os.path.join(out_dir,"variant_mapped_gene_details")
    #variant_gene_id_path = os.path.join(out_dir,"variant_mapped_gene_ids.txt")
    #make_file('\n'.join(deepcopy(genes)),variant_gene_id_path)
    gene_symbol_dict,gene_def_dict,results = convert_to_ensg(genes)

    for var in all_variants:
        for gene in var.genes:
            print("\n",gene)
            ## make sure there was a good symbol mapping
            print(gene_symbol_dict[gene])
            ###################################################
            ######## TODO: reference ensembl api to fetch these
            if gene_symbol_dict[gene] is None:
                temp_symbol_list = [gene]
            elif gene_symbol_dict[gene] == []:
                temp_symbol_list = [gene]
            elif gene_symbol_dict[gene] == [None]:
                temp_symbol_list = [gene]
            else:
                temp_symbol_list = deepcopy(gene_symbol_dict[gene])
            ## log the reverse lookup
            for symbol in temp_symbol_list:
                var.symbol_to_gene_lookup[symbol] = gene
            ## get the output text
            temp_symbol = "|".join(temp_symbol_list)
            ## set variant's symbols
            var.gene_symbols += temp_symbol_list
            
            ## make sure that there was an appropriate mapping
            print(gene_def_dict[gene])
            if gene_def_dict[gene] is None:
                temp_def = "None"
            elif gene_def_dict[gene] == []:
                temp_def = "None"
            elif gene_def_dict[gene] == [None]:
                temp_def = "None"
            else:
                temp_def = "|".join(deepcopy(gene_def_dict[gene]))
            var.detailed_gene_info.append([var.id, gene, temp_symbol, temp_def])

    #run_cmd("pyminer_gprofiler_converter.py -annotations -ID_list "+variant_gene_id_path+" -out "+annotation_out_path)
    return(all_variants)

def process_comention_table(comention_table):
    """
    Takes in the comention table & returns a dictionary with gene symbol as key and the list of results for all search terms as the values
    """
    ## has header
    comention_table = comention_table[1:]
    symbol_lookup = {line[0]:[] for line in comention_table}
    for i in range(len(comention_table)):
        temp_symbol = comention_table[i][0]
        temp_list = symbol_lookup[temp_symbol]
        temp_list.append(comention_table[i])
        symbol_lookup[temp_symbol] = temp_list
    return(symbol_lookup)
        

def get_best_var_search_term(gene_search_term_results, temp_id = ""):
    if len(gene_search_term_results) == 0:
        leading_candidate = [[temp_id]]
    else:
        default_best = 0
        leading_candidate = [gene_search_term_results[0][0]]
        for i in range(1,len(gene_search_term_results[0])):
            leading_candidate.append("NA")
        for line in gene_search_term_results:
            if line[-2] > default_best:## second to last entry is the best score for the given gene search-term pair
                default_best = line[-2]
                leading_candidate = deepcopy(line)
    return(leading_candidate)


def do_variant_to_search_term_mapping(all_variants, search_term, no_search_quote = False, out_dir=None):
    ## first prep the variant to ensg
    genes = []
    for var in all_variants:
        genes += var.gene_symbols
    genes = sorted(list(set(genes)))
    try:
        comention_summary_table = get_comention_summary(genes, search_term, no_search_quote = no_search_quote, new_pubmed = True)## reminder that this has a header
    except:
        print("We ran into problems scraping data from the new version of PubMed, we'll try the analysis again using the old version.")
        comention_summary_table = get_comention_summary(genes, search_term, no_search_quote = no_search_quote, new_pubmed = False)## reminder that this has a header
    header = ['variant','is_nearest_gene','gene'] + comention_summary_table[0]
    symbol_to_comention_dict = process_comention_table(comention_summary_table)
    for var in all_variants:
        for symbol in var.gene_symbols:
            var.gene_search_term_comentions += symbol_to_comention_dict[symbol]
    
    output = [header]
    output_best_only = [header]
    for var in all_variants:
        print(var)
        for i in range(len(var.gene_search_term_comentions)):
            line = var.gene_search_term_comentions[i]
            symbol = line[0]## this is the location that contains the symbol (ie. first column)
            #print(line)
            #print(symbol)
            #print(var.symbol_to_gene_lookup)
            ensg = var.symbol_to_gene_lookup[symbol]
            is_nearest_gene = ensg in var.nearest_genes
            line = [var.id, is_nearest_gene, ensg] + line
            #print("\t",line)
            var.gene_search_term_comentions[i] = line
            output.append(line)
        for line in var.gene_search_term_comentions:
            print(line)
        var.best_gene_search_term_match = get_best_var_search_term(var.gene_search_term_comentions, temp_id = var.id)
        output_best_only.append(var.best_gene_search_term_match)
    
    for var in all_variants:
        print(var.best_gene_search_term_match)

    if out_dir is not None:
        write_table(deepcopy(output),os.path.join(out_dir,"variant_to_gene_search_term_comention_summary.tsv"))
        write_table(deepcopy(output_best_only),os.path.join(out_dir,"variant_to_BEST_gene_search_term_comention_summary.tsv"))
    return(all_variants)


def get_most_significant_eQTL(full_eqtl_table, out_dir):
    out_table = [full_eqtl_table[0]]
    ## go through and log all of the variants and 
    combo_ids = []
    for i in range(1,len(full_eqtl_table)):
        combo_ids.append(full_eqtl_table[i][0]+full_eqtl_table[i][2])
    combo_best_p_dict = {combo:0 for combo in combo_ids}
    ## now catelogue the greatest negative log 10 p-value
    for i in range(1,len(full_eqtl_table)):
        temp_id = full_eqtl_table[i][0]+full_eqtl_table[i][2]
        try:
            float(full_eqtl_table[i][3])
        except:
            pass
        else:
            if float(full_eqtl_table[i][3]) > combo_best_p_dict[temp_id]:
                combo_best_p_dict[temp_id] = float(full_eqtl_table[i][3])
    ## now log them in the new output file
    for i in range(1,len(full_eqtl_table)):
        temp_id = full_eqtl_table[i][0]+full_eqtl_table[i][2]
        try:
            float(full_eqtl_table[i][3])
        except:
            pass
        else:
            if float(full_eqtl_table[i][3]) == combo_best_p_dict[temp_id]:
                out_table.append(full_eqtl_table[i])
    out_eqtl_path = os.path.join(out_dir,'var_to_all_gene_BEST_eQTL_summary.tsv')
    write_table(out_table, out_eqtl_path)
    return()


def do_eqtl_mapping(all_variants, out_dir = None):
    for var in all_variants:
        var.get_eqtl_for_genes()
        #print(var.eqtls)
    if out_dir is not None:
        out_table = [['rsID','is_nearest_gene','ENSG','neg_log10_p_val','tissue','value']]
        for var in all_variants:
            for gene, list_of_tissue_eqtl_dicts in var.eqtls.items():
                if list_of_tissue_eqtl_dicts is None:
                    out_table.append([var.id, gene in var.nearest_genes, gene, "NA","NA","NA"])
                else:
                    if len(list_of_tissue_eqtl_dicts) == 0:
                        out_table.append([var.id, gene in var.nearest_genes, gene, "NA","NA","NA"])
                    else:
                        for tissue_level_dict in list_of_tissue_eqtl_dicts:
                            if ("minus_log10_p_value" in tissue_level_dict) 
                                  and ("tissue" in tissue_level_dict)
                                  and ("value"in tissue_level_dict):
                                out_table.append([var.id, gene in var.nearest_genes, gene, removeNonAscii(str(tissue_level_dict["minus_log10_p_value"])), removeNonAscii(str(tissue_level_dict["tissue"])), removeNonAscii(str(tissue_level_dict["value"]))])
                            else:
                                out_table.append([var.id, gene in var.nearest_genes, gene, "NA","NA","NA"])
        out_eqtl_path = os.path.join(out_dir,'var_to_all_gene_full_eQTL_summary.tsv')
        get_most_significant_eQTL(out_table,out_dir)
        write_table(out_table, out_eqtl_path)
    return(all_variants)


def do_var_to_ensg_mapping(variant_file,
                           out_dir,
                           species,
                           lib_dir,
                           annotation_file = None,
                           id_list_file = None,
                           search_terms = None,
                           no_search_quote = True):
    if species is not "hsapiens":
        sys.exit("we can only work with human species rsIDs right now.")
    out_dir = process_dir(out_dir)
    
    ## check if we're mapping back to the original scRNAseq IDs (this is optional)
    if id_list_file != None and annotation_file != None:
        map_back_to_original = True
    else:
        map_back_to_original = False
    
    ## check if we're doing the search term co-mention (This is optional)
    if search_terms is not None:
        do_search_term = True
    else:
        do_search_term = False

    all_variants = map_variants_to_genes(variant_file, stringdb_dir = lib_dir, out_dir = out_dir)
    
    ## get the details of the ensembl genes
    all_variants = get_gene_level_info(all_variants, out_dir)

    ## get eQTL information
    all_variants = do_eqtl_mapping(all_variants, out_dir = out_dir)

    if map_back_to_original:
        ## if we're given the id_list file, it means that we're going to map things back from the ensembl IDs to the original IDs, likely used by the scRNAseq reference
        id_list = read_file(id_list_file,'lines')
        ensg_to_original_id = convert_anno_file_to_ensg_lookup(annotation_file, id_list)
        var_to_orig_ids = map_variants_to_original_ids(all_variants, ensg_to_original_id)

    if do_search_term:
        all_variants = do_variant_to_search_term_mapping(all_variants, read_file(search_terms,'lines'), out_dir = out_dir, no_search_quote = no_search_quote)
    
    save_dict(deepcopy(all_variants),os.path.join(out_dir,"variants.pkl"))
    return


##############################################################
def pyminer_annotate_variants_to_genes_main(args):
    do_var_to_ensg_mapping(variant_file = args.variants,
                           out_dir = args.out_dir,
                           species = args.species,
                           lib_dir = args.lib_dir,
                           annotation_file = args.annotations,
                           id_list_file = args.id_list,
                           search_terms = args.search_terms,
                           no_search_quote = args.no_search_quote)

if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument("-variants",'-vars',  
        help="a text file containing rsIDs separated on new lines",
        type = str)

    parser.add_argument("-out_dir", '-out', '-o',
        help="the directory that we'll write the results to",
        type = str)

    parser.add_argument("-species", '-s',
        help="a gProfiler accepted species code. Dafault = 'hsapiens'. Note that we currently only accept human rsIDs",
        type = str,
        default = 'hsapiens')

    parser.add_argument("-annotations", "-annotation_file","-anno_dict","-anno","-a",
        help="The annotation table - usually annotations.tsv")

    parser.add_argument("-lib_dir",'-lib',  
        help="the directory that contains the tad file - often titled 'tad_dict.pkl'",
        type = str)

    parser.add_argument('-id_list','-ids',
        help="the file containing the row ids. Usually ID_list.txt",
        type = str)

    parser.add_argument("-search_terms",'-terms', "-st" ,
        help="if you want to do the gene-symbol & search term co-mention analysis, give the file in here",
        type = str)

    parser.add_argument("-no_search_quote",
        help="whether or not to include quotes for the search terms",
        action = 'store_true')

    args = parser.parse_args()
    pyminer_annotate_variants_to_genes_main(args)